from .core.entity import Level, TileType
from .core.vector import Vector


# This function is responsible for
# loading the game.
def load_game(game_data: str, game):
    manag = game.manager
    manag.load_mod(game_data)

# This function is responsible for
# initializing the game.
def init_game(game):
    manag = game.manager

    for routine in manag.loader.routines['init.game']:
        routine(game)


from typing import List, Type, Callable, Optional, Any, Iterator

import os
import sys
import importlib
import warnings
import pkg_resources



class PluginNotFoundError(ImportError):
    pass


class UnmetPluginDependenciesError(PluginNotFoundError):
    pass


class RoutineSet(object):
    def __init__(self, loader: 'ModLoader'):
        self.__loader = loader

    def __getattr__(self, name):
        if name == '__loader':
            return self.__loader

        try:
            return self.__loader._routines[name]

        except KeyError:
            raise LookupError("No such routine '{}' registered in this loader!".format(name))

    def __getitem__(self, group: str) -> Iterator[Callable]:
        for r in self:
            if r.GROUP:
                a = r.GROUP.split('.')
                b = group.split('.')

                a = a[:len(b)]

                if tuple(a) == tuple(b):
                    yield r

    def __iter__(self) -> Iterator[Callable]:
        return iter(self.__loader._routines.values())


class ModLoader(object):
    def __init__(self):
        self._routines = {} # type: Dict[str, Callable]
        self.templates = [] # type: List[Type['EntityTemplate']]
        self.systems = [] # type: List[Type['System']]
        self.component_types = [] # type: List[Type['Component']]
        self.tile_types = [] # type: List['TileType']
        self.loaded = set()
        self.routines = RoutineSet(self)

    def load(self, name: str, module_loaded: Callable) -> Any:
        if name in self.loaded:
            return

        self.loaded.add(name)
        return module_loaded(self)

    def load_all(self):        
        for ep in pkg_resources.iter_entry_points('yodine.plugin'):
            self.load(ep.name, ep.load())
            yield ep.name

    def load_one(self, name: str) -> Any:
        for ep in pkg_resources.iter_entry_points('yodine.plugin', name):
            return self.load(ep.name, ep.load())

    def apply(self, manager: 'Manager'):
        for s in self.systems:
            manager.add_system(s)

        for t in self.templates:
            manager.register_template(t)

        for ct in self.component_types:
            manager.register_component(ct)

        for tt in self.tile_types:
            manager.add_tile_type(tt)

        self.templates = [] # type: List[Type['EntityTemplate']]
        self.systems = [] # type: List[Type['System']]
        self.component_types = [] # type: List[Type['Component']]
        self.tile_types = [] # type: List['TileType']

    def system_type(self, systype: Type['System']) -> Type['System']:
        self.systems.append(systype)
        return systype

    def add_tile_type(self, tiletype: 'TileType') -> 'TileType':
        self.tile_types.append(tiletype)
        return tiletype

    def routine(self, group: Optional[str] = None, name: Optional[str] = None) -> Callable:
        def _decorator(function: Callable) -> Callable:
            function.NAME = name
            function.GROUP = group

            self._routines[name or function.__name__] = function
            return function

        return _decorator

    def template(self, template: Type['EntityTemplate']) -> Type['EntityTemplate']:
        self.templates.append(template)
        return template

    def component_type(self, component_type: Type['Component']) -> Type['Component']:
        self.component_types.append(component_type)
        return component_type


import sqlite3
import uuid
import collections
import traceback
import warnings
import math
import json
import pyglet
import typing
import itertools

from typing import Optional, Iterable, Any, Tuple, TypeVar, Type, Iterator, Union, Callable, Dict
from .extension import ModLoader
from .vector import Vector, ComponentVector
from pyglet import gl



CIRCUM_MUL = 2 / math.sqrt(2)

ComponentType = Type['Component']



class EventContext(object):
    def __init__(self, name, manager, source, local = None, **kwargs):
        self.name = name
        self.manager = manager
        self.game = manager.game
        self.source = source

        self.local = local or self.game.id

        for attr_name, attr_value in kwargs.items():
            setattr(self, attr_name, attr_value)


class EntityContainer(object):
    def __init__(self):
        self.entity_component_index = {} # dict[entity => dict[component name => A implements Component]]
        self.entity_ids = [] # list<entity id>
        self.entity_id_set = set() # set<entity id>
        self.entity_count = 0

    def get_entities(self):
        return EntityContainer.__iter__(self)

    def get_component_types(self):
        return self.component_types

    def __len__(self):
        return self.entity_count

    def __iter__(self) -> Iterator['Entity']:
        return (Entity(self, eid) for eid in self.entity_ids)

    def load_entity(self, components: Optional[Iterable[Tuple[str, Any]]] = (), identifier: Optional[str] = None) -> 'Entity':
        if not identifier:
            identifier = str(uuid.uuid4())
            
        e = Entity(self, identifier)

        getattr(self, 'manager', self).register_delta('crt', identifier, getattr(self, 'id', None))

        for c in components:
            e.create_component(*c)
            
        getattr(self, 'manager', self).emit(e, 'loaded')

        return e

    def create_entity(self, components: Optional[Iterable[Tuple[str, Any]]] = (), identifier: Optional[str] = None) -> 'Entity':
        e = self.load_entity(components, identifier)

        getattr(self, 'manager', self).emit_all('spawn', e)
        getattr(self, 'manager', self).emit(e, 'spawned')

        return e

    def create_templated_entity(self, template_name: str, components: Optional[Iterable[Tuple[str, Any]]] = (), identifier: Optional[str] = None) -> 'Entity':
        e = self.templates[template_name]
        ent = e.spawn(self, components, identifier)
        
        return ent

    def remove_entity(self, e: 'Entity') -> None:
        if e.level is self:
            e.remove()


# === Level data ===

class TileType(object):
    name = None # type: str
    post = False

    def __init__(self, sprite: pyglet.resource.image, name: Optional[str] = None):
        self.image = sprite.get_texture()
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)   

        size = 35

        self.image.width = size
        self.image.height = size
        self.image.anchor_x = size / 2
        self.image.anchor_y = size / 2

        self.sprite_cache = {} # type: Dict[Tuple[int, int], pyglet.sprite.Sprite]
        self.post_sprite_cache = {} # type: Dict[Tuple[int, int], pyglet.sprite.Sprite]

        if name:
            self.name = name

        else:
            self.name = type(self).name

    def unset(self, level: 'Level', x, y):
        if (x, y) in self.sprite_cache:
            self.sprite_cache[x, y][1].delete()
            del self.sprite_cache[x, y]

        if (x, y) in self.post_sprite_cache:
            self.post_sprite_cache[x, y][1].delete()
            del self.post_sprite_cache[x, y]

    def force_render_tile(self, window: pyglet.window.Window, level: 'Level', wx, wy):
        pyglet.sprite.Sprite(self.image, wx + 35 / 2, wy + 35 / 2).draw()

    def _render(self, window: pyglet.window.Window, level: 'Level', x, y, wx, wy, cam_angle, cam_zoom, batch = None, cache = None):
        if self.image:
            cache = cache if cache is not None else self.sprite_cache

            if (x, y) in cache:
                ow, sprite = cache[x, y]
                sprite.visible = True

                try:
                    sprite.rotation = math.degrees(cam_angle)
                    sprite.scale = cam_zoom
            
                    if (wx, wy) != ow:
                        sprite.position = (wx, wy)

                except AttributeError as err:
                    print("WARNING: Bad tile sprite found at x={},y={} (see error below); deleting and ignoring.".format(x, y))
                    traceback.print_exc()

                    if (x, y) in cache:
                        del cache[x, y]

                    return

            else:
                sprite = pyglet.sprite.Sprite(self.image, wx, wy, batch=batch or level.batch)
                cache[x, y] = ((wx, wy), sprite)

    def render(self, window: pyglet.window.Window, level: 'Level', x, y, wx, wy, cam_angle, cam_zoom):
        if not self.post:
            self._render(window, level, x, y, wx, wy, cam_angle, cam_zoom)

    def post_render(self, window: pyglet.window.Window, level: 'Level', x, y, wx, wy, cam_angle, cam_zoom):
        if self.post:
            self._render(window, level, x, y, wx, wy, cam_angle, cam_zoom, level.post_batch, self.post_sprite_cache)
        
    def unrender(self, window: pyglet.window.Window, level: 'Level', x, y, wx, wy, cam_angle, cam_zoom):
        if (x, y) in self.sprite_cache:
            _, s = self.sprite_cache[x, y] # type: pyglet.sprite.Sprite
            s.visible = False

    def post_unrender(self, window: pyglet.window.Window, level: 'Level', x, y, wx, wy, cam_angle, cam_zoom):
        if (x, y) in self.post_sprite_cache:
            _, s = self.post_sprite_cache[x, y] # type: pyglet.sprite.Sprite
            s.visible = False

    def tick(self, manager: 'Manager', *args):
        pass

    def is_inside(self, manager: 'Manager', ent: 'Entity') -> bool:
        if 'position' not in ent:
            return False

        vec = ComponentVector(ent['position'])
        x = math.floor(vec.x / 35)
        y = math.floor(vec.y / 35)
        
        return manager.current_level.tiles.get((x, y), None) == self.name
        
    def _on(self, event_name: str, event: EventContext, *args, **kwargs):
        if hasattr(self, 'on_' + event_name):
            return getattr(self, 'on_' + event_name)(event, *args, **kwargs)

class Trigger(object):
    def __init__(self, level: EntityContainer, x, y, width = 1, height = 1):
        self.level = level
        self.x = x
        self.y = y
        self.width = width
        self.height = height

        self.inside = set()

    def is_inside(self, ent: 'Entity') -> bool:
        if 'position' not in ent:
            return False

        left = self.x * 35
        right = (self.x + self.width) * 35
        top = self.y * 35
        bottom = (self.y + self.height) * 35

        vec = ComponentVector(ent['position'])
        
        return vec.x >= left and vec.x <= right and vec.y >= top and vec.y <= bottom

    def tick(self, *args):
        for e in self.level.manager:
            if self.is_inside(e):
                if e not in self.inside:
                    self.inside.add(e)
                    self.triggered(e)

            elif e in self.inside:
                self.inside.remove(e)
    


# === Entity Component System ===



class Level(EntityContainer):
    def __init__(self, lid  : str, manager: 'Manager'):
        EntityContainer.__init__(self)
    
        self.id = lid
        self.manager = manager
        self.tiles = {}
        self.deltas = []

        self.batch = pyglet.graphics.Batch()
        self.post_batch = pyglet.graphics.Batch()
        self.last_rendered = set()

    def __iter__(self) -> Iterator['Entity']:
        for e in self.get_entities():
            yield e

        for e in self.manager.get_entities():
            yield e

    def emit(self, source, event_name, *args):
        return self.manager.emit(source, event_name, *args)

    def create_templated_entity(self, template_name: str, components: Optional[Iterable[Tuple[str, Any]]] = (), identifier: Optional[str] = None) -> 'Entity':
        e = self.manager.templates[template_name]
        ent = e.spawn(self, components, identifier)
        
        return ent

    def get_component_types(self):
        return self.manager.get_component_types()

    def entities_at(self, x: int, y: int) -> Iterator['Entity']:
        left = x * 35
        right = (x + 1) * 35

        top = y * 35
        bottom = (y + 1) * 35

        for e in self.get_entities():
            if 'position' in e:
                pos = ComponentVector(e['position'])

                if pos.x >= left and pos.x <= right and pos.y >= top and pos.y <= bottom:
                    yield e

                del pos

    def tiles(self, tt: TileType) -> Iterator[Vector]:
        for xy, tile in self.tiles.items():
            if tile == tt.name:
                yield Vector(xy)

    def transform_position(self, vec: Vector) -> Vector:
        return type(vec)(int(vec.x / 35), int(vec.y / 35))

    def rectangle(self, start: Vector, width: int, height: int, tile: str):
        for y in range(int(start.y), int(start.y) + height):
            for x in range(int(start.x), int(start.x) + width):
                if (x, y) in self.tiles and self.tiles[x, y] != tile:
                    self.manager.tile_types[self.tiles[x, y]].unset(self, x, y)

                self.tiles[x, y] = tile

        self.deltas.append(('rect', int(start.x), int(start.y), width, height, tile))

    def set(self, pos: Vector, tile: str):
        if self.tiles[int(pos.x), int(pos.y)] == tile:
            return

        if (int(pos.x), int(pos.y)) in self.tiles:
            self.manager.tile_types[self.tiles[int(pos.x), int(pos.y)]].unset(self, int(pos.x), int(pos.y))

        self.tiles[int(pos.x), int(pos.y)] = tile
        self.deltas.append(('set', int(pos.x), int(pos.y), tile))

    def save(self):
        return json.dumps(self.get_save())

    def get_save(self):
        return {
            'deltas': self.deltas
        }

    def load(self, data: str):
        self.load_save(json.loads(data))

    def load_save(self, save):
        self.deltas = []

        deltas = list(save['deltas'])

        self.tiles = {}

        for i, d in enumerate(deltas):
            if d not in deltas[i + 1:]:
                self.apply_delta(d)

    def apply_delta(self, d):
        kind = d[0]

        if kind == 'set':
            self.set(Vector((d[1], d[2])), *d[3:])

        elif kind == 'rect':
            self.rectangle(Vector((d[1], d[2])), *d[3:])

        else:
            raise ValueError("Unknown level delta type: " + repr(kind))

    def render(self, window: pyglet.window.Window):
        #sw = self.manager.camera_transform(window, 0, 0)
        #xdirx, xdiry = self.manager.camera_transform_delta(window, 35, 0)
        #ydirx, ydiry = self.manager.camera_transform_delta(window, 0, 35)

        for (x, y), tile in self.tiles.items():
            if tile:
                t = self.manager.tile_types[tile]

                if not t.post:
                    #wx, wy = sw
                    #wx += xdirx * x + xdiry * y
                    #wy += ydirx * x + ydiry * y

                    wx, wy = self.manager.camera_transform(window, x * 35 + 35 / 2, y * 35 + 35 / 2)
                    
                    circumscribed = 35 * CIRCUM_MUL * self.manager.camera_zoom

                    if wx > -circumscribed and wx < window.width + circumscribed * 2 and wy > -circumscribed and wy < window.height + circumscribed * 2:
                        t.render(window, self, x, y, wx, wy, self.manager.camera_angle, self.manager.camera_zoom)
                        self.last_rendered.add((x, y))

                    elif (x, y) in self.last_rendered:
                        t.unrender(window, self, x, y, wx, wy, self.manager.camera_angle, self.manager.camera_zoom)

        self.batch.draw()

    def post_render(self, window: pyglet.window.Window):
        #sw = self.manager.camera_transform(window, 0, 0)
        #xdirx, xdiry = self.manager.camera_transform_delta(window, 35, 0)
        #ydirx, ydiry = self.manager.camera_transform_delta(window, 0, 35)

        for (x, y), tile in self.tiles.items():
            if tile:
                t = self.manager.tile_types[tile]

                if t.post:
                    #wx += xdirx * x + xdiry * y
                    #wy += ydirx * x + ydiry * y
                    wx, wy = self.manager.camera_transform(window, x * 35 + 35 / 2, y * 35 + 35 / 2)

                    circumscribed = 35 * CIRCUM_MUL * self.manager.camera_zoom

                    if wx > -circumscribed and wx < window.width + circumscribed and wy > -circumscribed and wy < window.height + circumscribed:
                        t.post_render(window, self, x, y, wx, wy, self.manager.camera_angle, self.manager.camera_zoom)
                        self.last_rendered.add((x, y))

                    elif (x, y) in self.last_rendered:
                        t.post_unrender(window, self, x, y, wx, wy, self.manager.camera_angle, self.manager.camera_zoom)

        self.post_batch.draw()

    def __del__(self):
        for (x, y), tile in self.tiles.items():
            self.manager.tile_types[tile].unset(self, x, y)

        del self.tiles


class Manager(EntityContainer):
    def __init__(self, game, window):
        EntityContainer.__init__(self)

        self.game = game # type; yodine.game.Game
        self.window = window # type: pyglet.window.Window
        #self.entity_list = []
        #self.entity_index = {}

        self.systems = [] # type: List[System]
        self.event_listeners = {} # type: Dict[str, Callable]
        self.templates = {}
        self.tile_types = {} # type: Dict[str, TileType]
        self.component_types = component_types
        self.camera = Vector((0, 0))
        self.levels = {} # type: Dict[str, Level]
        self.default_level = self.create_level('_DEFAULT')
        self.current_level = self.default_level # type: Level
        self.loader = ModLoader()
        self.global_event_listeners = set()
        self.camera_angle = 0
        self.camera_zoom = 1.2
        self.dtime = None

        self.set_middle()

    def __contains__(self, eid: str) -> bool:
        if eid in self.entity_id_set:
            return True

        if self.current_level and eid in self.current_level.entity_id_set:
            return True

        return False

    def update(self, other_manager: 'Manager'):
        for e in self:
            loaded = e.id not in other_manager.entity_ids
            e.copy_transfer(other_manager)

            if loaded:
                other_manager.emit(e, 'loaded')

    def register_delta(self, delta_name, *args):
        if self.game.server:
            na = []

            for d in self.game.server.change_accum:
                dn2 = d[0]

                if delta_name == dn2:
                    if args == d[1:]:
                        continue

                    elif delta_name == 'set':
                        eid1, cname1, cval1 = d[1:]
                        eid2, cname2, cval2 = args
                        
                        if eid1 == eid2 and cname1 == cname2:
                            continue

                na.append(d)

            na.append((delta_name, *args))

            del self.game.server.change_accum
            self.game.server.change_accum = na
    
    def set_middle(self, size_x = None, size_y = None):
        if (size_x and size_y) or self.window:
            self.middle = Vector(((size_x or self.window.width) / 2, (size_y or self.window.height) / 2))

        else:
            self.middle = Vector((0, 0))

    def find_entity(self, eid):
        if eid in self.entity_ids:
            return Entity(self, eid)

        elif eid in self.current_level.entity_ids:
            return Entity(self.current_level, eid)

        raise ValueError("Entity not found: {}".format(eid))

    def camera_transform(self, window, x, y):
        v = Vector((x - self.camera.x, y - self.camera.y)).rotate(-self.camera_angle)

        v.x = v.x * self.camera_zoom + self.middle.x
        v.y = v.y * self.camera_zoom + self.middle.y

        return v

    def camera_transform_delta(self, window, x, y):
        return Vector((x, y)).rotate(-self.camera_angle) * self.camera_zoom

    def force_render_tile(self, window: pyglet.window.Window, tiletype: str, wx, wy):
        self.tile_types[tiletype].force_render_tile(self.current_level, window, wx, wy)

    def un_camera_transform(self, window, x, y):
        v = Vector((x - self.middle.x, y - self.middle.y)).rotate(self.camera_angle)

        v.x = v.x / self.camera_zoom + self.camera.x
        v.y = v.y / self.camera_zoom + self.camera.y

        return v

    def __iter__(self) -> Iterator['Entity']:
        for e in self.get_entities():
            yield e

        if self.current_level is not None:
            for e in self.current_level.get_entities():
                yield e

    def all_entity_ids(self) -> Iterator['str']:
        for eid in self.entity_id_set:
            yield eid

        if self.current_level is not None:
            for eid in self.current_level.entity_id_set:
                yield eid

    def change_level(self, lid: str):
        self.current_level = self.levels[lid]
        self.register_delta('clv', lid)

    def add_level_save(self, lid: str, save) -> Level:
        l = Level(lid, self)
        l.load_save(save)

        return self.add_level(l)

    def add_level(self, level: Level) -> Level:
        self.levels[level.id] = level

        if level.id != '_DEFAULT':
            self.register_delta('nlv', level.id, level.get_save())
            
        return level

    def create_level(self, lid: str) -> Level:
        return self.add_level(Level(lid, self))

    def add_tile_type(self, tt: TileType):
        self.tile_types[tt.name] = tt

    def move_camera(self, x=0, y=0, angle=0):
        self.camera += Vector((x, y))
        self.camera_angle += angle

    def set_camera(self, x=None, y=None, angle=None, zoom=None):
        if x is not None:
            self.camera.x = x

        if y is not None:
            self.camera.y = y

        if angle is not None:
            self.camera_angle = angle

        if zoom is not None:
            self.camera_zoom = zoom

    def load_mod(self, plugin_name: str):
        self.loader.load_one(plugin_name)

        for r in self.loader.routines['preload.' + plugin_name]:
            r(self.game)

        self.loader.apply(self)

        for r in self.loader.routines['postload.' + plugin_name]:
            r(self.game)

    def apply_mod(self, plugin_name: str, func: Callable):
        self.loader.load(plugin_name, func)

        for r in self.loader.routines['preload.' + plugin_name]:
            r(self.game)

        self.loader.apply(self)

        for r in self.loader.routines['postload.' + plugin_name]:
            r(self.game)

    def load_all_mods(self):
        for plugin_name in self.loader.load_all():
            for r in self.loader.routines['preload.' + plugin_name]:
                r(self.game)

            self.loader.apply(self)

            for r in self.loader.routines['postload.' + plugin_name]:
                r(self.game)

    def register_template(self, template: Type['EntityTemplate']) -> Type['EntityTemplate']:
        self.templates[template.name] = template(self)
        return template

    def register_component(self, cotype: Type['Component']) -> Type['Component']:
        self.component_types[cotype.__name__] = cotype
        return cotype

    def tick(self, dtime: float):
        self.set_middle()
        self.dtime = dtime

        for e in self:
            for s in self.systems:
                if hasattr(s, 'tick'):
                    s._tick(e, dtime)

        for t in self.tile_types.values():
            t.tick(self, dtime)

    def render(self, window: pyglet.window.Window):
        window.clear()

        if self.current_level is not None:
            self.current_level.render(window)

        for e in self:
            for s in self.systems:
                if hasattr(s, 'render'):
                    s._render(e, window)

        if self.current_level is not None:
            self.current_level.post_render(window)

    def apply(self, window: pyglet.window.Window) -> pyglet.window.Window:
        @window.event
        def on_draw():
            self.render(window)

        return window

    def reset_systems(self):
        self.systems = []

    def add_system(self, s: Type['System']) -> None:
        self.systems.append(s(self))

    def iter_grouped_templates(self, template_group: str) -> Iterator[Type['EntityTemplate']]:
        for template in self.templates:
            a = template.group.split('.')
            b = template_group.split('.')

            a = a[:len(b)]

            if tuple(a) == tuple(b):
                yield template

    def listen(self, event_name):       
        check_name = '_'.join(event_name) if type(event_name) is tuple else event_name

        def _decorator(self, func):
            if check_name not in self.event_listeners:
                self.event_listeners[check_name] = set()
            
            self.event_listeners[check_name] |= {func}
            return func

        return _decorator

    def add_listener(self, event_name, func):
        if event_name not in self.event_listeners:
            self.event_listeners[event_name] = set()

        self.event_listeners[event_name] |= {func}

    def global_listener(self, func):
        self.global_event_listeners |= {func}

    def emit_local(self, source, event_name, local, *args, **kwargs):
        check_name = '_'.join(event_name) if type(event_name) is tuple else event_name

        if check_name in self.event_listeners:
            for func in self.event_listeners[event_name]:
                func(source, *args)


        event_ctx = EventContext(event_name, self, source, local, **kwargs)

        for system in self.systems:
            system._on(check_name, event_ctx, *args)

        for tt in self.tile_types.values():
            tt._on(check_name, event_ctx, *args)

        for gel in self.global_event_listeners:
            gel(event_ctx, event_name, *args)

    def emit_all_local(self, event_name, local, *args, **kwargs):
        for e in self:
            self.emit_local(e, event_name, local, *args, **kwargs)

    def emit(self, source, event_name, *args, **kwargs):
        self.emit_local(source, event_name, self.game.id, *args, **kwargs)

    def emit_all(self, event_name, *args, **kwargs):
        self.emit_all_local(event_name, self.game.id, *args, **kwargs)

    def __getitem__(self, entity_id: str) -> 'Entity':
        return self.find_entity(entity_id)


component_types = {}

def register_component(cotype):
    component_types[cotype.__name__] = cotype
    return cotype


@register_component
class Component(object):
    def __init__(self, entity: 'Entity', name: str, value: Any = None):
        self.entity = entity
        self.name = name
        self._value = value
    
    def force_set(self, value: Any = None) -> bool:
        if self._value == value:
            return False

        self._value = value
        return True

    def set(self, value: Any = None):
        if self.force_set(value):
            self.entity.manager.register_delta('set', self.entity.id, self.name, value)
            self.entity.manager.emit(self.entity, ('change', self.name), self)

    def __repr__(self):
        return '[{} {} = {}]'.format(type(self).__name__, self.name, repr(self.value))

    def __getattr__(self, name):
        if name == 'value':
            return self.get()

        return super().__getattribute__(name)

    def __setattr__(self, name, value: Any = None):
        if name == 'value':
            self.set(value)

        else:
            super().__setattr__(name, value)

    def get(self):
        return self._value


class Entity(object):
    def __init__(self, level: 'EntityContainer', identifier = None):
        self.level = level
        self.manager = getattr(level, 'manager', level)
        self.id = identifier or str(uuid.uuid4())

        self.level.entity_component_index.setdefault(self.id, {})

        if self.id not in self.level.entity_id_set:
            self.level.entity_ids.append(self.id)
            self.level.entity_id_set.add(self.id)
            self.level.entity_count += 1

    def transfer(self, new_level: EntityContainer):
        if self.id in new_level.entity_id_set or new_level is self.level:
            #raise RuntimeError("Tried to transfer an entity to a container it is already at!")
            return

        self.remove()
        self.copy_transfer(new_level)

        self.manager.register_delta('tsf', self.id, new_level.id)

    def copy_transfer(self, new_level: EntityContainer):
        comps = list(self.get_components()) # type: List[Component]
        
        new_level.entity_component_index[self.id] = {}

        if self.id not in new_level.entity_id_set:
            new_level.entity_ids.append(self.id)
            new_level.entity_id_set.add(self.id)
            new_level.entity_count += 1
        
        for cmpnt in comps:
            new_level.entity_component_index[self.id][cmpnt.name] = cmpnt

    def get_components(self) -> Iterator['Component']:
        return self.level.entity_component_index[self.id].values()

    def __hash__(self):
        return hash(self.id)

    def remove(self):
        del self.level.entity_component_index[self.id]
        self.level.entity_id_set.remove(self.id)
        self.level.entity_ids.remove(self.id)
        self.level.entity_count -= 1

        self.manager.register_delta('dsp', self.id)

    def create_component(self, name: str, value = None, kind = None) -> 'Component':
        ct = self.level.get_component_types()

        if kind is None:
            kind = Component

        elif kind in ct:
            kind = ct[kind]

        elif isinstance(kind, str):
            raise ValueError("Unknown component type: " + repr(kind))

        elif not (isinstance(kind, type) and issubclass(kind, Component)):
            raise TypeError("Bad component type (expected str, Type[Component], None): " + repr(kind))

        comp = kind(self, name, value)
        self.level.entity_component_index[self.id][name] = comp

        self.manager.register_delta('mkc', self.id, name, value, kind.__name__)

        return comp

    def __iter__(self):
        return iter(self.get_components())

    def __getitem__(self, comp_name: str) -> 'Component':
        return self.level.entity_component_index[self.id][comp_name]

    def __setitem__(self, comp_name: str, value: Optional[Any] = None):
        if comp_name in self:
            c = self[comp_name]
            c.value = value

        else:
            self.create_component(comp_name, value)

    def __contains__(self, comp_name: str) -> bool:
        return comp_name in self.level.entity_component_index[self.id]

    def __delitem__(self, comp_name: str):
        if comp_name not in self:
            return

        del self.level.entity_component_index[self.id][comp_name]

        if comp_name != 'localplayer':
            self.manager.emit(self, 'component_remove', comp_name)

        self.manager.register_delta('del', self.id, comp_name)

    def has(self, *comp_names: Iterable[str]) -> bool:
        return all(cn in self for cn in comp_names)

    def has_any(self, *comp_names: Iterable[str]) -> bool:
        return any(cn in self for cn in comp_names)

    def __repr__(self):
        return '<[ {} {} ]>'.format(self.id, ', '.join(repr(c) for c in self.get_components()))


class System(object):
    listeners = []
    component_types = []
    component_checks = {}
    component_defaults = {}

    def component_check(self, func):
        def _inner(ctx: Union[EventContext, Entity], *args):
            source = getattr(ctx, 'source', ctx)

            if not self.component_types:
                return func(source, *args)

            else:
                components = {}

                for cotype, cocheck in self.component_checks.items():
                    if cotype not in source or source[cotype].value != cocheck:
                        return False

                for cotype in self.component_types:
                    if cotype in source:
                        components[cotype] = source[cotype]

                    else:
                        return False

                for cotype, codefault in self.component_defaults.items():
                    if cotype not in source:
                        source[cotype] = codefault

                    components[cotype] = source[cotype]

                func(ctx, *args, **components)
                return True

        return _inner

    def __init__(self, manager: Manager):
        self.manager = manager
        self.game = manager.game
        
        self.listeners = type(self).listeners
        self.component_types = type(self).component_types
        self.component_checks = type(self).component_checks
        self.component_defaults = type(self).component_defaults

        for event_name, func in self.listeners:
            manager.init_event(event_name)
            manager.add_listener(event_name, func)

        self.system_init()

    def system_init(self):
        pass
    
    def _tick(self, entity: Entity, *args) -> None:
        self.component_check(self.tick)(entity, *args)

    def _render(self, entity: Entity, window: pyglet.window.Window, *args) -> None:
        self.component_check(self.render)(entity, window, *args)

    def _on(self, event_name: str, event: EventContext, *args):
        if hasattr(self, 'on_' + event_name):
            return self.component_check(getattr(self, 'on_' + event_name))(event, *args)

    #def tick(self, entity: Entity, *args, **kwargs) -> None:
    #    pass

    #def render(self, entity: Entity, window: pyglet.window.Window, *args, **kwargs) -> None:
    #    pass


class SystemRegistry(object):
    def __init__(self):
        self.system_types = []

    def define(self, sys: Type[System]) -> Type[System]:
        self.system_types.append(sys)

        return sys

    def apply(self, manager: Manager) -> None:
        for st in self.system_types:
            manager.add_system(st)


class EntityTemplate(object):
    name = None # type: str
    group = None # type: Optional[str]
    default_components = [] # type: Iterable[Tuple[str, Any]]

    def __init__(self, manager: Manager):
        self.manager = manager
        self.default_component_index = {
            v[0]: (v[1] if len(v) > 1 else None) for v in self.default_components
        }

    def get_value(self, cname):
        return self.default_component_index[cname]

    def spawn(self, level: Level, components: Optional[Iterable[Tuple[str, Any]]] = (), identifier: Optional[str] = None) -> Entity:
        components = list(components)
        has_components = set(c[0] for c in components)
        
        for dc in self.default_components:
            if dc[0] not in has_components:
                components.append(dc)

        return level.create_entity(components, identifier)



all_systems = SystemRegistry()



from .entity import Manager, Component, Entity
from .vector import Vector
from typing import Callable

import time
import uuid
import socket
import trio
import pyglet

try:
    import simplejson as json

except ImportError:
    import json



class MultiplayerError(BaseException):
    pass

class MultiplayerProtocolError(MultiplayerError):
    pass

class MultiplayerUserError(MultiplayerError):
    pass


def json_adv_process(arg, name_map):
    if isinstance(arg, list):
        return [json_adv_process(a, name_map) for a in arg]

    elif isinstance(arg, dict):
        return {name_map[k]: json_adv_process(v, name_map) for k, v in arg.items()}

    return arg

def json_names(arg, res):
    if isinstance(arg, list):
        for a in arg:
            json_names(a, res)

    elif isinstance(arg, dict):
        for k, v in arg.items():
            res.add(k)
            json_names(v, res)

letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
letter_combos = list(letters) + [a + b for a in letters for b in letters]

def json_adv_dump(arg):
    names = set()
    json_names(arg, names)
    names = list(names)
    name_map = {n: letter_combos[i] for i, n in enumerate(names)}

    return json.dumps([{letter_combos[i]: n for i, n in enumerate(names)}, json_adv_process(arg, name_map)])

def json_adv_load(arg):
    name_map, arg = json.loads(arg)
    return json_adv_process(arg, name_map)

def dump_arg(arg):
    if isinstance(arg, Vector):
        return { 't': 'vec', 'coord': [arg.x,arg.y] }

    elif isinstance(arg, Component):
        return { 't': 'cmp', 'eid': arg.entity.id, 'name': arg.name, 'value': arg.value }

    elif isinstance(arg, Entity):
        return { 't': 'ent', 'eid': arg.id }

    elif isinstance(arg, pyglet.window.BaseWindow):
        return { 't': 'wnd' }

    elif isinstance(arg, list):
        return { 't': 'lst', 'values': [dump_arg(a) for a in arg] }

    elif isinstance(arg, tuple):
        return { 't': 'tpl', 'values': [dump_arg(a) for a in arg] }

    elif isinstance(arg, dict):
        return { 't': 'dct', 'items': { k: dump_arg(v) for k, v in arg.items() } }

    return { 't': '', 'value': arg }


class Client(object):
    def __init__(self, game: '..game.Game', server_addr):
        self.game = game
        self.manager = game.manager
        self.client = socket.create_connection(server_addr)
        self.client.setblocking(False)
        self.addr = server_addr
        self.players = {}
        self.events = []
        self.writes = []
        self.buffer = b''
        self.outbound = b''
        self.running = True
        self.emit_callbacks = {}

        self.average_latency = 0

    def send(self, command, *args):
        self._write('\x00'.join(
            [command, *(json_adv_dump(dump_arg(a)) for a in args)]
        ).encode('utf-8') + b'\n')

    def emit(self, source, name, *args, **kwargs):
        broad_id = str(uuid.uuid4())
        self.send('BROADCAST', broad_id, source.id, name, args, kwargs)
        when_sent = time.time()

        def _inner(when_received, *args):
            if self.average_latency == 0:
                self.average_latency = when_received

            else:
                self.average_latency = (when_received - when_sent + self.average_latency * 4) / 5

        self.emit_callbacks[broad_id] = _inner

    def stop(self):
        self._write(b'LEAVE\n')
        self.client.close()

        self.running = False

    def load_arg(self, arg):
        if arg['t'] == 'vec':
            return Vector(arg['coord'])

        elif arg['t'] == '':
            return arg['value']

        elif arg['t'] == 'cmp':
            e = self.manager.find_entity(arg['eid'])

            if arg['name'] in e:
                c = e[arg['name']]
                c.force_set(arg['value'])

            else:
                c = Component(e, arg['name'], arg['value'])
                e.level.entity_component_index[e.id][c.name] = c

            return c

        elif arg['t'] == 'ent':
            return self.manager.find_entity(arg['eid'])

        elif arg['t'] == 'wnd':
            return self.manager.window

        elif arg['t'] == 'lst':
            return [self.load_arg(a) for a in arg['values']]

        elif arg['t'] == 'tpl':
            return tuple([self.load_arg(a) for a in arg['values']])

        elif arg['t'] == 'dct':
            return { k: self.load_arg(v) for k, v in arg['items'].items() }
        
        else:
            raise ValueError("Unknown JSON argument type: {}".format(repr(arg['t'])))

    def process_writes(self):
        nw = []

        for when, data in self.writes:
            self.outbound += data

        self.writes = nw

    def _write(self, data):
        #self.outbound += data
        self.writes.append((time.time(), data))

    async def run(self):
        self.running = True        
        self.buffer = b''

        commands = {}

        def command(name):
            def _decorator(func):
                commands[name.upper()] = func
                
                return func
            return _decorator

        self.ignore_events_until = 0

        @command('DELTA')
        def receive_delta(*changes):
            # print('Received changes!')

            for change in changes:
                args = change[1:]

                if change[0] == 'set':
                    eid, cname, cval = args

                    if eid in self.manager:
                        self.manager[eid][cname] = cval

                elif change[0] == 'mkc':
                    eid, cname, cval, ckind = args

                    if eid in self.manager:
                        self.manager[eid].create_component(cname, cval, ckind)

                elif change[0] == 'del':
                    eid, cname = args

                    if eid in self.manage:
                        e = self.manager[eid]
                        
                        if cname in e:
                            del e[cname]

                elif change[0] == 'crt':
                    eid, level = args

                    if level is None:
                        e = self.manager.create_entity([], eid)

                    else:
                        e = self.manager.levels[level].create_entity([], eid)

                elif change[0] == 'dsp':
                    eid, = args

                    if eid in self.manager:
                        self.manager[eid].remove()

                elif change[0] == 'tsf':
                    eid, level = args

                    if eid in self.manager:
                        if level is None:
                            self.manager[eid].transfer(self.manager)

                        else:
                            self.manager[eid].transfer(self.manager.levels[level])

                elif change[0] == 'clv':
                    level, = args

                    self.manager.change_level(level)

                elif change[0] == 'nlv':
                    level, save = args

                    if level not in self.manager.levels:
                        self.manager.add_level_save(level, save)
                
            print('{} changes now'.format(len(changes)).ljust(40), end='\r')

        #@command('EVENT')
        #def receive_broadcast_event(when, local, source, event_name, args, kwargs):
        #    if when <= self.ignore_events_until:
        #        return
        #
        #    if time.time() - when < 1 / 4:
        #        self.manager.emit_local(source, event_name, local, *args, **kwargs)
        #
        #    elif time.time() - when > 2:
        #        self.ignore_events_until = time.time()
        #        self.send('SNAPSHOT')

        @command('ERR')
        def got_error(code, err):
            category = str(code)[0]
            msg = 'Status code {}: {}'.format(code, err)
            
            if category == '1':
                raise MultiplayerProtocolError(msg)

            elif category == '2':
                raise MultiplayerUserError(msg)

            else:
                raise MultiplayerError(msg)

        @command('GOT_BROADCAST')
        def broadcast_acknowledged(b_id, *args):
            if b_id in self.emit_callbacks:
                self.emit_callbacks[b_id](*args)

        @command('INIT_LEVELS')
        def init_levels(start, *levels):
            del self.manager.levels
            self.manager.levels = {}

            for v in levels:
                lid = v['lid']
                assert 'deltas' in v

                self.manager.add_level_save(lid, v)

            self.manager.change_level(start)

            print('Received levels.')

        @command('INIT_ENTITIES')
        def init_entities(*entities):
            for e in self.manager:
                e.remove()

            for v in entities:
                eid = v['eid']
                level = v.get('level', None)
                components = v['components']

                if level:
                    e = self.manager.levels[level].create_entity(components, eid)

                else:
                    e = self.manager.create_entity(components, eid)

            print('Received {} entities.'.format(len(entities)))

        @command('RNG')
        def set_rng_state(*state):
            self.game.random.setstate(state)
            print('Received RNG state.')

        async def _recv(data):
            self.buffer += data
            lines = self.buffer.split(b'\n')
            self.buffer = lines[-1]

            for data in lines[:-1]:
                try:
                    data = data.decode('utf-8')

                except UnicodeDecodeError:
                    print(data)
                    raise

                keycode = data.split('\x00')[0]
                values = [self.load_arg(json_adv_load(a)) for a in data.split('\x00')[1:]]

                if keycode.upper() in commands:
                    commands[keycode.upper()](*values)

        print('Joining...')
        self.send('JOIN', self.game.id, self.game.player_name)

        while self.events:
            e = self.events.pop(0)
            self.send_event(e)
        
        self.events = None

        async with trio.open_nursery() as nursery:
            while self.running:
                sleep_amount = 1 / 30
                data = b''

                try:
                    data = self.client.recv(8192)

                    if data == b'':
                        self.stop()
                        return

                    else:
                        nursery.start_soon(_recv, data)

                except ConnectionResetError:
                    self.stop()
                    return

                except BlockingIOError:
                    sleep_amount = 1 / 15

                self.process_writes()

                try:
                    sent = self.client.send(self.outbound)

                except BlockingIOError:
                    pass

                else:
                    self.outbound = self.outbound[sent:]

                    if self.outbound or len(data) == 8192:
                        sleep_amount = 1 / 60
                    
                await trio.sleep(sleep_amount)

            self.outbound = b''
            

class Server(object):
    def emit(self, source, event_name, *args, **kwargs):
        for sess in self.clients.keys():
            self.send(sess, 'EVENT', self.game.id, source, event_name, args, kwargs)

    def load_arg(self, arg):
        if arg['t'] == 'vec':
            return Vector(arg['coord'])

        elif arg['t'] == '':
            return arg['value']

        elif arg['t'] == 'cmp':
            e = self.manager.find_entity(arg['eid'])

            if arg['name'] in e:
                c = e[arg['name']]
                c.force_set(arg['value'])

            else:
                c = Component(e, arg['name'], arg['value'])
                e.level.entity_component_index[e.id][c.name] = c

            return c

        elif arg['t'] == 'ent':
            return self.manager.find_entity(arg['eid'])

        elif arg['t'] == 'wnd':
            return self.manager.window

        elif arg['t'] == 'lst':
            return [self.load_arg(a) for a in arg['values']]

        elif arg['t'] == 'tpl':
            return tuple([self.load_arg(a) for a in arg['values']])

        elif arg['t'] == 'dct':
            return { k: self.load_arg(v) for k, v in arg['items'].items() }
        
        else:
            raise ValueError("Unknown JSON argument type: {}".format(repr(arg['t'])))

    def send(self, session, command, *args):
        self._write(session, '\x00'.join([
            command,
            *(json_adv_dump(dump_arg(a)) for a in args),
        ]).encode('utf-8') + b'\n')

    def send_deltas(self):
        self.last_t = time.time()

        if self.change_accum:
            for s in self.players.keys():
                #for chg in self.change_accum:
                #    print('delta sent:   ', *chg)

                self.send(s, 'DELTA', *self.change_accum)

            self.change_accum = []

    def __init__(self, game: '..game.Game', port: int, player_type: str = 'player'):
        self.game = game
        self.manager = game.manager
        self.player_names = set()
        self.clients = {}
        self.players = {}
        self.outbound = {}
        self.writes = {}
        self.locals = {}
        self.change_accum = []
        self.port = int(port)
        self.player_type = player_type
        self.running = False

        self.buffer = ''

    def _write(self, s, data):
        self.writes[s].append((time.time(), data))

    async def accept(self, client: socket.socket, address):
        client.setblocking(0)

        session = str(uuid.uuid4())
        self.outbound[session] = b''
        self.writes[session] = []

        self.players[session] = None
        self.clients[session] = None

        commands = {}

        def _write(data):
            self._write(session, data)

        def error(*args):
            self.send(session, 'ERR', *args)

        print('New client connected from:', ':'.join(str(x) for x in address))

        async def _recv(data):
            try:
                lines = (_recv.buffer + data.decode('utf-8')).split('\n')

            except ValueError:
                #_write(b'ERR\x00102\x00malformed command line buffering' + b'\n')
                error(102, 'malformed command line buffering')
                return

            _recv.buffer = lines[-1]

            for command in lines[:-1]:
                command = command.strip()

                try:
                    name = command.split('\x00')[0]

                except ValueError:
                    #_write(b'ERR\x00103\x00malformed command arguments' + b'\n')
                    error(103, 'malformed command arguments')

                for target_name, func in commands.items():
                    if name.upper() == target_name.upper():
                        #print('  -  Executing command:', target_name.upper())
                        values = []

                        for val in command.split('\x00')[1:]:
                            try:
                                values.append(self.load_arg(json_adv_load(val)))

                            except BaseException:
                                error(101, 'bad JSON data')
                                #_write(b'ERR\x00101\x00bad JSON data' + b'\n')

                                print(command.replace('\x00', ' \x1B[1m|\x1B[0m '))
                                return

                            except BaseException as err:
                                print(command.replace('\x00', ' \x1B[1m|\x1B[0m '))
                                raise
                        
                        func(*values)

        _recv.buffer = ''

        def command(target_name: str):
            def _decorator(func: Callable):
                commands[target_name] = func
                return func
                                
            return _decorator

        def send_entities():
            loaded = []

            for entity in self.manager:
                comp = [(comp.name, comp.value, type(comp).__name__ if type(comp) is not Component else None) for comp in entity.get_components()]

                e = {
                    'eid': entity.id,
                    'components': comp
                }

                if entity.level is not self.manager:
                    e['level'] = entity.level.id

                loaded.append(e)

            self.send(session, 'INIT_ENTITIES', *loaded)
            print('Sent entities')
        
        @command('join')
        def on_join(local_id, name):
            if self.players[session]:
                print('/!\ Already joined:', name)
                error(201, 'already joined')

            elif name in self.player_names:
                print('/!\ Name taken:', name)
                error(203, 'name taken')

            else:
                p = None

                for e in self.manager:
                    if e.has('name', 'localplayer') and e['name'].value == name:
                        p = e
                        p['localplayer'] = local_id
                        break

                lvls = []

                for lid, level in self.manager.levels.items():
                    lvls.append({
                        'lid': lid,
                        'deltas': level.deltas
                    })
                    
                self.send(session, 'INIT_LEVELS', self.manager.current_level.id, *lvls)
                send_entities()
                self.send(session, 'RNG', *self.game.random.getstate())

                if not p:
                    lvl = self.manager.current_level

                    if not lvl:
                        lvl = self.manager

                    print('Creating player...')
                    p = lvl.create_templated_entity(self.player_type, [('name', name), ('localplayer', local_id)])

                eid = p.id

                self.players[session] = p
                self.locals[session] = local_id

                print('Joined:', name, '(local: {}, id: {})'.format(local_id, eid))

                # Force resending deltas
                if time.time() - self.last_t > update_interval:
                    self.send_deltas()

        @command('snapshot')
        def on_request_snapshot():
            if session in self.players:
                send_entities()

        @command('leave')
        def on_leave():
            if self.players[session] is None:
                error(202, 'tried to leave a game the client is already not connected to', "You're just like the kid who's not invited to the party, but comes anyway, and takes a dump in the pool just for the reactions.")
                print('/!\ Not in the game', address)

            else:
                print('Left:', self.players[session]['name'].value)
                self.players[session].remove()

        @command('broadcast')
        def on_event(broadcast_id, source, event_name, args, kwargs):
            local = self.locals[session]
            
            for sess in self.clients.keys():
                if sess != session:
                    self.send(sess, 'EVENT', time.time(), local, source, event_name, args, kwargs)

            if source not in self.manager.all_entity_ids():
                error(204, 'no such entity')

            self.manager.emit_local(self.manager[source], event_name, local, *args, **kwargs)
            self.send(session, 'GOT_BROADCAST', broadcast_id, time.time())

        update_interval = 1 / 15
        self.last_t = time.time()

        async with trio.open_nursery() as nursery:
            while self.running:
                sleep_amount = 1 / 30
                data = b''

                try:
                    data = client.recv(8192)

                    if data == b'':
                        del self.clients[session]

                        if session in self.players: del self.players[session]
                        if session in self.locals: del self.locals[session]

                        return

                    else:
                        nursery.start_soon(_recv, data)

                except (ConnectionResetError, BrokenPipeError):
                    del self.clients[session]

                    if session in self.players: del self.players[session]
                    if session in self.locals: del self.locals[session]
                            
                    return

                except BlockingIOError:
                    sleep_amount = 1 / 20

                if time.time() - self.last_t > update_interval:
                    self.send_deltas()
                    
                self.process_writes(session)

                try:
                    sent = client.send(self.outbound[session])

                except (ConnectionResetError, BrokenPipeError):
                    del self.clients[session]

                    if session in self.players: del self.players[session]
                    if session in self.locals: del self.locals[session]

                    return

                except BlockingIOError:
                    pass

                else:
                    self.outbound[session] = self.outbound[session][sent:]

                    if self.outbound or len(data) == 8192:
                        sleep_amount = 1 / 60
                    
                await trio.sleep(sleep_amount)
        
    async def run(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(('0.0.0.0', self.port))
        self.socket.listen(5)
        self.socket.setblocking(False)

        self.running = True

        async with trio.open_nursery() as nursery:
            while self.running:
                try:
                    (conn, addr) = self.socket.accept()

                except BlockingIOError:
                    pass

                else:
                    nursery.start_soon(self.accept, conn, addr)

                await trio.sleep(0.5)

        self.clients = {}

    def stop(self):
        self.running = False
        self.socket.close()

        for s in self.clients.values():
            if s:
                s.close()

    def process_writes(self, s):
        nw = []

        for when, data in self.writes[s]:
            #if time.time() - when > 5:
            #    continue

            self.outbound[s] += data

        self.writes[s] = nw


import sys, functools
import random
import trio
import os
import math
import pyglet
import pyglet.gl as pgl
import pyglet.window.mouse as pgmouse

from . import game
from .core.vector import Vector, ComponentVector
from .core.trig import fast_sin
from .core.entity import Entity, System, EntityTemplate, EventContext
from .core.extension import ModLoader
from pyglet.window import key
from tkinter import *
from tkinter import ttk
from typing import Optional

try:
    import simplejson as json

except ImportError:
    import json



class EditorToolkit(Frame):
    def __init__(self, editor, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.editor = editor
        self.grid(column=0, row=0, sticky=(N, W, E, S))

        self.init_widgets()

        for child in self.winfo_children():
            child.grid_configure(padx=5, pady=5)

    def add_component(self):
        nam = self.component_name.get()
        val = self.component_value.get()

        try:
            self.editor.extra_components[nam] = (json.loads(val) if val != '' else None)

        except json.JSONDecodeError:
            self.error.set('Component value is not JSON parseable!')

        else:
            self.component_name.set('')
            self.component_value.set('')
            self.error.set('')
            
            ttk.Label(self.component_list, text=nam, justify='right', font='Helvetica 12 bold').grid(row=len(self.editor.extra_components) - 1, column=0, sticky=(E,))

            e = ttk.Entry(self.component_list, text=val, justify='left')
            e.insert(0, val)
            e.configure(state='readonly')
            e.grid(row=len(self.editor.extra_components) - 1, column=1, sticky=(W,))

    def init_widgets(self):
        self.component_name = StringVar()
        self.component_value = StringVar()
        self.error = StringVar()
        
        cname = ttk.Entry(self, textvariable=self.component_name)
        cval = ttk.Entry(self, textvariable=self.component_value)

        ttk.Label(self, text='Component Name', justify='center').grid(row=0, column=0, sticky=(S,))
        ttk.Label(self, text='Component Value', justify='center').grid(row=0, column=1, sticky=(S,))

        cname.grid(row=1, column=0, sticky=(N,))
        cval.grid(row=1, column=1, sticky=(N,))

        button = ttk.Button(self, text='Add Component', command=self.add_component)
        button.grid(row=2, column=1, sticky=(N, W))

        error_lbl = ttk.Label(self, textvariable=self.error, justify='right', foreground='red')
        error_lbl.grid(row=2, column=0, sticky=(N, S, E))

        self.component_list = Frame(self, background='#DDD', height=350)
        self.component_list.grid(column=0, columnspan=2, row=3, rowspan=3, sticky='nswe')

        self.component_list.columnconfigure(0, weight=1)
        self.component_list.rowconfigure(0, weight=1)
        self.component_list.grid_propagate(0)


class Editor(game.Game):
    def __init__(self, game_name: str, database_file: Optional[str] = None):
        super().__init__(game_name, database_file)

        self.root = Tk()  
        self.root.title('Yoded Tools')
        self.frame = EditorToolkit(self, self.root, background='white')

        self.root.columnconfigure(0, weight=1)
        self.root.rowconfigure(0, weight=1)
        self.root.protocol("WM_DELETE_WINDOW", self.stop)
        self.running = False

        self.extra_components = {}

    def init_player(self):
        pass

    def stop(self):
        super().stop()
        self.running = False

    async def tk_loop(self):
        self.running = True

        while self.running:
            self.root.update_idletasks()
            self.root.update()
            
            await trio.sleep(1 / 15)

        self.root.destroy()

    async def loop(self):
        async with trio.open_nursery() as nursery:
            nursery.start_soon(super().loop)
            nursery.start_soon(self.tk_loop)


def editor_plugin(loader: ModLoader):
    @loader.template
    class EditorCursor(EntityTemplate):
        name = 'edcursor'
        group = 'overlays'

        default_components = [
            ('position', (0, 0)),
            ('cam', (0, 0)),
            ('shown', False),
            ('age', 0),
            ('clickable', (35, 35)),
            ('position', (0, 0)),
            ('edcursor',),
            ('panspeed', 150),
        ]

    @loader.template
    class TilePaletteItem(EntityTemplate):
        name = 'edpaletteitem'
        group = 'overlays'

        default_components = [
            ('palette',),
            ('index', 0),
            ('tiletype',),
            ('clickable', (35, 35)),
            ('position', (0, 0)),
            ('selected', False),
            ('age', 0),
        ]

    
    @loader.template
    class ObjectPaletteItem(EntityTemplate):
        name = 'edobjpalette'
        group = 'overlays'

        default_components = [
            ('palette',),
            ('index', 0),
            ('template',),
            ('clickable', (35, 35)),
            ('position', (0, 0)),
            ('selected', False),
            ('age', 0),
        ]


    @loader.system_type
    class S_NormalRender(System):
        component_types = ['position', 'sprite']
        component_defaults = {'angle': 0}
        component_checks = {'render': 'normal'}

        def system_init(self):
            self.images = {}
            self.sprites = {}

            for r in self.manager.loader.routines['get.yodine.sprites']:
                self.images.update(r())

        def on_spawned(self, event: EventContext, sprite, **kwargs):
            entity = event.source
            self.sprites[entity.id] = pyglet.sprite.Sprite(self.images[sprite.value])
            self.sprites[entity.id].rotation = -90

        def on_loaded(self, event: EventContext, sprite, **kwargs):
            entity = event.source
            self.sprites[entity.id] = pyglet.sprite.Sprite(self.images[sprite.value])
            self.sprites[entity.id].rotation = -90

        def on_change_sprite(self, event: EventContext, new_sprite, **kwargs):
            entity = event.source

            if entity.id in self.sprites:
                self.sprites[entity.id].delete()

            self.sprites[entity.id] = pyglet.sprite.Sprite(self.images[new_sprite.value])
            self.sprites[entity.id].rotation = -90

        def render(self, entity: Entity, window: pyglet.window.Window, position, angle, sprite, **kwargs):
            if entity.id not in self.sprites:
                self.sprites[entity.id] = pyglet.sprite.Sprite(self.images[sprite.value])

            sprite = self.sprites[entity.id] # type: pyglet.sprite.Sprite

            vec = ComponentVector(position)
            newvec = Vector(entity.manager.camera_transform(window, vec.x, vec.y))

            sprite.x, sprite.y = newvec.x, newvec.y
            sprite.scale = self.manager.camera_zoom
            
            if 'localplayer' in entity:
                c_angle = -self.manager.camera_angle

            else:
                c_angle = self.manager.camera_angle

            sprite.rotation = -90 + math.degrees(angle.value + c_angle)

            if 'radius' in entity:
                rad = entity['radius'].value
                sprite.scale /= max(sprite.image.width, sprite.image.height) / rad

            if 'scale' in entity:
                scale = entity['scale'].value
                sprite.scale *= scale

            sprite.draw()



    def get_pos(entity: Entity, component_name: str = 'position') -> ComponentVector:
        assert component_name in entity
        return ComponentVector(entity[component_name])

    @loader.system_type
    class S_LabelRender(System):
        component_types = ['name', 'position']

        def render(self, entity: Entity, window: pyglet.window.Window, name, position, *args, **kwargs) -> None:
            pos = get_pos(entity)
            x, y = self.manager.camera_transform(window, *pos)
            y += 75

            label = pyglet.text.Label(
                name.value,
                anchor_x = 'center',
                anchor_y = 'center',
                x = x,
                y = y,
                bold = True
            )
            label.draw()


    @loader.system_type
    class S_EditorCursor(System):
        component_types = ['edcursor', 'position', 'shown', 'age', 'cam', 'panspeed']

        def tick(self, entity: Entity, dtime: float, age, cam, position, panspeed, shown, *args, **kwargs):
            self.keyboard = self.manager.game.keyboard

            age.value += dtime
            
            psp = panspeed
            panspeed = panspeed.value * dtime
            
            cam = ComponentVector(cam)
            position = ComponentVector(position)

            if self.keyboard[key.LSHIFT] or self.keyboard[key.RSHIFT]:
                shown.value = False
            
            else:
                shown.value = True

            if self.keyboard[key.UP]:
                cam.y += panspeed
                position.y += panspeed

            elif self.keyboard[key.DOWN]:
                cam.y -= panspeed
                position.y -= panspeed

            if self.keyboard[key.LEFT]:
                cam.x -= panspeed
                position.x -= panspeed

            elif self.keyboard[key.RIGHT]:
                cam.x += panspeed
                position.x += panspeed

            if self.keyboard[key.G]:
                psp.value += 10 * dtime

            if self.keyboard[key.F]:
                psp.value -= 10 * dtime
                psp.value = max(psp.value, 0.025)

            if self.keyboard[key.NUM_ADD]:
                entity.manager.camera_zoom += 0.2 * dtime
                entity.manager.camera_zoom = min(entity.manager.camera_zoom, 4)

            elif self.keyboard[key.NUM_SUBTRACT]:
                entity.manager.camera_zoom -= 0.2 * dtime
                entity.manager.camera_zoom = max(entity.manager.camera_zoom, 0.125)

            entity.manager.set_camera(cam.x, cam.y, 0)
            
        def on_mouse_move(self, event: EventContext, window, new_pos: Vector, pos_diff: Vector, position, shown, **kwargs):
            entity = event.source
            self.mouse_move(entity, window, new_pos, pos_diff, position, shown)

        def on_mouse_drag(self, event: EventContext, window, new_pos: Vector, pos_diff: Vector, _, _1, position, shown, **kwargs):
            entity = event.source
            self.mouse_move(entity, window, new_pos, pos_diff, position, shown)

        def mouse_move(self, entity: Entity, window, new_pos: Vector, pos_diff: Vector, position, shown, **kwargs):
            position = ComponentVector(position)

            nx, ny = entity.manager.un_camera_transform(window, new_pos.x, new_pos.y)

            position.x = nx
            position.y = ny

        def on_mouse_enter(self, event: EventContext, window, xy, shown, **kwargs):
            if self.keyboard and (self.keyboard[key.LSHIFT] or self.keyboard[key.RSHIFT]):
                shown.value = False
            
            else:
                shown.value = True

        def on_mouse_leave(self, event: EventContext, window, xy, shown, **kwargs):
            shown.value = False

        def render(self, entity: Entity, window: pyglet.window.Window, position, shown, age, **kwargs) -> None:
            if shown.value:
                position = ComponentVector(position)
                cx, cy = position.x, position.y

                wx = math.floor(cx / 35) * 35
                wy = math.floor(cy / 35) * 35

                wx, wy = entity.manager.camera_transform(window, wx, wy)

                wx = int(wx)
                wy = int(wy)

                color = (
                    255,
                    255,
                    20,
                    int((fast_sin(age.value * math.pi) + 1) * 255 / 2)
                )
                size = int(35 * self.manager.camera_zoom)

                pyglet.graphics.draw(4, pgl.GL_QUADS,
                    ('v2i', (
                        # position
                        wx,        wy,
                        wx + size, wy,
                        wx + size, wy + size,
                        wx,        wy + size,
                    )),

                    ('c4B', color * 4)
                )


    @loader.system_type
    class S_Clickables(System):
        component_types = ['clickable', 'position']
        component_defaults = {'down': 0}

        def inside(self, mousepos, position, clickable):
            pos = ComponentVector(position)
            cli = ComponentVector(clickable)

            return (
                mousepos.x > pos.x and mousepos.x <= pos.x + cli.x and
                mousepos.y > pos.y and mousepos.y <= pos.y + cli.y
            )

        def on_mouse_release(self, event, window, pos, button, modifiers, clickable, position, down):
            entity = event.source
            down = down.value

            if self.inside(pos, position, clickable):
                self.manager.emit(entity, 'click', pos, button, modifiers)
                down &= ~button

            entity['down'] = down


        def on_mouse_press(self, event, window, pos, button, modifiers, clickable, position, down):
            entity = event.source
            down = down.value

            if self.inside(pos, position, clickable):
                down |= button

            entity['down'] = down
        

        def render(self, entity: Entity, window: pyglet.window.Window, down, clickable, position, **kwargs) -> None:
            if down.value:
                position = ComponentVector(position)
                size = ComponentVector(clickable)

                color = (
                    40,
                    40,
                    40,
                    70,
                )

                pyglet.graphics.draw(4, pgl.GL_QUADS,
                    ('v2i', (
                        # position
                        int(position.x),          int(position.y),
                        int(position.x + size.x), int(position.y),
                        int(position.x + size.x), int(position.y + size.y),
                        int(position.x),          int(position.y + size.y),
                    )),

                    ('c4B', color * 4)
                )


    @loader.system_type
    class S_ObjectPalette(System):
        component_types = ['palette', 'index', 'template', 'selected', 'position', 'age', 'sprite']

        def system_init(self):
            self.selected = None
            self.images = {}
            self.sprites = {}
            self.spawned = []

            for r in self.manager.loader.routines['get.yodine.sprites']:
                self.images.update(r())

        def on_key_release(self, event, window,   button, modifiers,   selected, **kwargs):
            entity = event.source

            if selected.value and button == key.U and self.spawned:
                e = self.spawned.pop()
                e.remove()

        def on_spawned(self, event: EventContext, sprite, **kwargs):
            entity = event.source
            self.sprites[entity.id] = pyglet.sprite.Sprite(self.images[sprite.value])

        def on_loaded(self, event: EventContext, sprite, **kwargs):
            entity = event.source
            self.sprites[entity.id] = pyglet.sprite.Sprite(self.images[sprite.value])

        def on_change_sprite(self, event: EventContext, new_sprite, **kwargs):
            entity = event.source

            if entity.id in self.sprites:
                self.sprites[entity.id].delete()

            self.sprites[entity.id] = pyglet.sprite.Sprite(self.images[new_sprite.value])

        def tick(self, entity: Entity, dtime: float, age, *args, **kwargs):
            age.value += dtime

        def on_resetobjpalselect(self, event, **kwargs):
            entity = event.source
            entity['selected'].value = False

        def on_mouse_release(self, event, window,    xy, button, modifiers,    template, selected,   **kwargs):
            entity = event.source
            if button & pgmouse.LEFT and (modifiers & pyglet.window.key.MOD_SHIFT):
                self.put(entity, window, xy, template, selected)

        def on_click(self, event,    xy, button, modifiers,    template, selected,   **kwargs):
            entity = event.source
            if button & pgmouse.RIGHT:
                self.manager.emit_all('resetobjpalselect')
                entity['selected'].value = True
                self.selected = template.value

        def put(self, entity, window, xy, template, selected):
            if self.selected is not None and selected.value:
                tx, ty = self.manager.un_camera_transform(window, xy.x, xy.y)

                self.spawned.append(self.manager.current_level.create_templated_entity(self.selected, [
                    ('position', [tx, ty]),
                    *list(self.manager.game.extra_components.items())
                ]))

                self.manager.game.extra_components = {}

                for child in self.manager.game.frame.component_list.winfo_children():
                    child.grid_remove()
        
        def render(self, entity: Entity, window,   index, template, selected, position, age, sprite,   **kwargs):
            if entity.id not in self.sprites:
                self.sprites[entity.id] = pyglet.sprite.Sprite(self.images[sprite.value])

            sprite = self.sprites[entity.id] # type: pyglet.sprite.Sprite
                
            wx = 10
            wy = window.height - 40 - index.value * 45

            pos = ComponentVector(position)
            pos.x, pos.y = wx, wy

            wx = int(wx)
            wy = int(wy)

            sprite.x, sprite.y = wx + 35 / 2, wy + 35 / 2
            sprite.scale = 35 / max(sprite.image.width, sprite.image.height)

            if 'scale' in entity:
                scale = entity['scale'].value
                sprite.scale *= scale

            sprite.draw()

            if selected.value:
                color = (
                    60,
                    230,
                    20,
                    int((fast_sin(age.value * math.pi) + 1) * 255 / 2)
                )

                pyglet.graphics.draw(4, pgl.GL_QUADS,
                    ('v2i', (
                        wx,      wy,
                        wx + 35, wy,
                        wx + 35, wy + 35,
                        wx,      wy + 35,
                    )),
                    ('c4B', color * 4)
                )
            

    @loader.system_type
    class S_TilePalette(System):
        component_types = ['palette', 'index', 'tiletype', 'selected', 'position', 'age']

        def system_init(self):
            self.selected = None
            self.start_pos = None
            self.curr_pos = None

        def tick(self, entity: Entity, dtime: float, age, *args, selected, **kwargs):
            age.value += dtime

        def on_resetpalselect(self, event, **kwargs):
            entity = event.source
            entity['selected'].value = False

        def on_mouse_release(self, event, window,    xy, button, modifiers,    tiletype, selected,   **kwargs):
            entity = event.source

            if button & pgmouse.LEFT and not (modifiers & (pyglet.window.key.MOD_SHIFT | pyglet.window.key.MOD_CTRL)) and not self.start_pos:
                self.put(entity, window, xy, tiletype, selected)

            if self.start_pos:
                if self.selected is not None and selected.value:
                    self.put_rectangle()
                            
                    self.start_pos = None
                    self.curr_pos = None

        def on_mouse_press(self, event, window,   xy, button, modifiers,  tiletype, selected,    **kwargs):
            entity = event.source

            if modifiers & pyglet.window.key.MOD_CTRL:
                tp = self.tile_pos(window, xy)
                self.start_pos = tp
                self.curr_pos = tp

        def on_mouse_drag(self, event, window,   xy, dxy, button, modifiers,   tiletype, selected,   **kwargs):
            entity = event.source

            if button & pgmouse.LEFT and not (modifiers & (pyglet.window.key.MOD_SHIFT | pyglet.window.key.MOD_CTRL)):
                self.put(entity, window, xy, tiletype, selected)

            elif self.start_pos:
                tp = self.tile_pos(window, xy)

                if modifiers & pyglet.window.key.MOD_CTRL:
                    self.curr_pos = tp

                else:
                    if self.selected is not None and selected.value:
                        self.put_rectangle()
                        
                        self.start_pos = None
                        self.curr_pos = None

        def tile_pos(self, window, vec):
            tx, ty = self.manager.un_camera_transform(window, *vec)

            tx = math.floor(tx / 35)
            ty = math.floor(ty / 35)

            return (tx, ty)

        def put_rectangle(self):
            x1, y1 = self.start_pos
            x2, y2 = self.curr_pos

            if x1 > x2:
                x2, x1 = x1, x2

            if y1 > y2:
                y2, y1 = y1, y2

            self.manager.current_level.rectangle(Vector((x1, y1)), x2 - x1 + 1, y2 - y1 + 1, self.selected)

        def on_click(self, event,    xy, button, modifiers,    tiletype, selected,   **kwargs):
            entity = event.source

            if button & pgmouse.RIGHT:
                self.manager.emit_all('resetpalselect')
                entity['selected'].value = True
                self.selected = tiletype.value

        def put(self, entity, window, xy, tiletype, selected):
            if self.selected is not None and selected.value:
                tx, ty = self.tile_pos(window, xy)

                if (tx, ty) not in self.manager.current_level.tiles or self.manager.current_level.tiles[tx, ty] != self.selected:
                    self.manager.current_level.set(Vector((tx, ty)), self.selected)
        
        def render(self, entity: Entity, window,   index, tiletype, selected, position, age,   **kwargs):
            wx = window.width - 40
            wy = window.height - 40 - index.value * 45
            entity.manager.force_render_tile(window, tiletype.value, wx, wy)
            position.value = [wx, wy]

            if selected.value:
                color = (
                    20,
                    60,
                    255,
                    int((fast_sin(age.value * math.pi) + 1) * 255 / 2)
                )

                pyglet.graphics.draw(4, pgl.GL_QUADS,
                    ('v2i', (
                        wx,      wy,
                        wx + 35, wy,
                        wx + 35, wy + 35,
                        wx,      wy + 35,
                    )),

                    ('c4B', color * 4)
                )

                if self.start_pos:
                    sp = self.manager.camera_transform(window, *(Vector(self.start_pos) * 35)).ints()
                    cp = (self.manager.camera_transform(window, *(Vector(self.curr_pos) * 35))).ints()

                    if sp.x > cp.x:
                        sp.x, cp.x = cp.x, sp.x

                    if sp.y > cp.y:
                        sp.y, cp.y = cp.y, sp.y

                    cp += self.manager.camera_transform_delta(window, 35, 35)
                    cp = cp.ints()

                    color = (
                        200,
                        75,
                        75,
                        int((fast_sin(age.value * math.pi) + 1) * 255 / 2)
                    )

                    pyglet.graphics.draw(4, pgl.GL_QUADS,
                        ('v2i', (
                            sp.x, sp.y,
                            cp.x, sp.y,
                            cp.x, cp.y,
                            sp.x, cp.y,
                        )),

                        ('c4B', color * 4)
                    )

    @loader.system_type
    class S_FPSCounter(System):
        component_types = ['fpscounter']

        def render(self, entity: Entity, window: pyglet.window.Window, *args, **kwargs) -> None:
            if hasattr(self, 'dtime'):
                label = pyglet.text.Label(
                    str(round(1 / self.dtime, 1)),
                    x = 30,
                    y = 30,
                    bold = True
                )
                label.draw()

        def tick(self, entity: Entity, dtime: float, *args, **kwargs):
            self.dtime = dtime

    @loader.system_type
    class S_SizeRandom(System):
        component_types = ['sizerandom']

        def on_spawned(self, event: EventContext, sizerandom, **kwargs):
            entity = event.source
            
            entity['radius'].value = entity['radius'].value + random.uniform(-entity['sizerandom'].value, entity['sizerandom'].value)


    @loader.system_type
    class S_Readable(System):
        component_types = ['message', 'position']

        def render(self, entity: Entity, window: pyglet.window.Window, position, message, **kwargs):
            pos = ComponentVector(position)
            dist = (self.manager.camera - pos).size()

            if dist < 150:
                alpha = 1 - dist / 150

                pos = get_pos(entity)
                x, y = self.manager.camera_transform(window, *pos)
                y += 60

                label = pyglet.text.Label(
                    message.value,
                    anchor_x = 'center',
                    anchor_y = 'center',
                    x = x,
                    y = y,
                    bold = True,
                    font_size=14 * self.manager.camera_zoom,
                    color=(190, 130, 30, int(255 * alpha)),
                    multiline=True,
                    width=window.width * 3 / 5,
                    align='center',
                )
                label.draw()


    @loader.system_type
    class S_AngleRandom(System):
        component_types = ['anglerandom']

        def on_spawned(self, event: EventContext, **kwargs):
            entity = event.source
            entity['angle'].value = random.uniform(0, math.pi * 2)

    @loader.routine('preload.Yoded')
    def editor_preinit(game: Editor):
        manag = game.manager
        manag.systems = []

    @loader.routine('postload.Yoded')
    def editor_init(game: Editor):
        manag = game.manager

        for e in manag:
            if 'edcursor' in e or 'palette' in e:
                e.remove()

        manag.create_templated_entity('edcursor', [
            ('cshown', True)
        ])

        for i, tt in enumerate(manag.tile_types.keys()):
            manag.create_templated_entity('edpaletteitem', [
                ('index', i),
                ('tiletype', tt)
            ])

        i = 0
        for tname, template in manag.templates.items():
            if hasattr(template, 'editor_visible') and template.editor_visible:
                manag.create_templated_entity('edobjpalette', [
                    ('index', i),
                    ('sprite', template.get_value('sprite')),
                    ('template', tname)
                ])
                i += 1

        print("Editor initialized successfully.")


def editor_main():
    editor = Editor(os.environ.get('YODINE_GAME', 'yodine_data'), sys.argv[1] if len(sys.argv) > 1 else 'main.save.json')

    if len(sys.argv) > 2:
        lid = sys.argv[2]

        if lid in editor.manager.levels:
            editor.manager.change_level(lid)

        else:
            editor.manager.create_level(lid)

    editor.manager.apply_mod('Yoded', editor_plugin)

    print('Starting Editor...')
    editor.run()


if __name__ == '__main__':
    editor_main()


import math
import struct

from .trig import fast_sin, fast_cos



class Vector(object):
    def __init__(self, x, y = None):
        if y:
            self.x = x
            self.y = y

        else:
            try:
                self.x, self.y = x

            except TypeError:
                raise ValueError("Non-vectorial value: {}".format(repr(x)))

    def __iter__(self):
        return iter((self.x, self.y))

    def __getitem__(self, co):
        if co in (0, 'x'):
            return self.x

        if co in (1, 'y'):
            return self.y

        raise KeyError("No such key in a vector: ", repr(co))

    def ints(self):
        return Vector((
            int(self[0]),
            int(self[1])
        ))

    def rotate(self, angle: float) -> 'Vector':
        c = math.cos(angle)
        s = math.sin(angle)

        return Vector((
            self.x * c - self.y * s,
            self.x * s + self.y * c,
        ))

    def size(self) -> float:
        return math.sqrt(self[0] ** 2 + self[1] ** 2)

    def sqsize(self) -> float:
        return float(self[0] ** 2 + self[1] ** 2)

    def fisrsize(self) -> float:
        y = self.sqsize()
        threehalfs = 1.5
        x2 = y * 0.5
    
        packed_y = struct.pack('f', y)       
        i = struct.unpack('i', packed_y)[0]  # treat float's bytes as int 
        i = 0x5f3759df - (i >> 1)            # arithmetic with magic number
        packed_i = struct.pack('i', i)
        y = struct.unpack('f', packed_i)[0]  # treat int's bytes as float
        
        y = y * (threehalfs - (x2 * y * y))  # Newton's method
        return y

    def unit(self) -> 'Vector':
        if self.sqsize() == 0:
            return Vector((0, 0))

        if self.sqsize() == 1:
            return self

        return self * self.fisrsize()

    def __add__(self, b: 'Vector') -> 'Vector':
        return Vector((self[0] + b.x, self[1] + b.y))

    def __neg__(self) -> 'Vector':
        return Vector((-self[0], -self[1]))

    def __mul__(self, b) -> 'Vector':
        try:
            return Vector((self[0] * b.x, self[1] * b.y))

        except AttributeError:
            return Vector((self[0] * b, self[1] * b))

    def __truediv__(self, b) -> 'Vector':
        try:
            return Vector((self[0] / b.x, self[1] / b.y))

        except AttributeError:
            return Vector((self[0] / b, self[1] / b))

    def dot(self, b: 'Vector') -> float:
        return self[0] * b.vec[0] + self[1] * b.vec[1]

    def __sub__(self, b: 'Vector') -> 'Vector':
        return self + (-b)

    def __repr__(self) -> str:
        return "{}(x={},y={})".format(type(self).__name__, self[0], self[1])


class ComponentVector(Vector):
    def __init__(self, component: 'Component'):
        self.component = component

    def __iadd__(self, b: 'Vector') -> 'ComponentVector':
        res = self + b
        self.x = res.x
        self.y = res.y
        return self

    def __isub__(self, b: 'Vector') -> 'ComponentVector':
        res = self - b
        self.x = res.x
        self.y = res.y
        return self

    def __imul__(self, b: 'Vector') -> 'ComponentVector':
        res = self * b
        self.x = res.x
        self.y = res.y
        return self

    def __itruediv__(self, b: 'Vector') -> 'ComponentVector':
        res = self / b
        self.x = res.x
        self.y = res.y
        return self

    def __lshift__(self, b: 'Vector') -> 'ComponentVector':
        self.x = b.x
        self.y = b.y
        return self

    @property
    def x(self):
        return self.component.value[0]

    @property
    def y(self):
        return self.component.value[1]

    @x.setter
    def x(self, val):
        self.component.value = [val, self.component.value[1]]

    @y.setter
    def y(self, val):
        self.component.value = [self.component.value[0], val]


import math



sin_cache = {}
cos_cache = {}


def fast_sin(x):
    deg = int(x * 100)

    if deg in sin_cache:
        return sin_cache[deg]

    else:
        res = math.sin(x)
        sin_cache[deg] = res

        return res

def fast_cos(x):
    deg = int(x * 100)

    if deg in cos_cache:
        return cos_cache[deg]

    else:
        res = math.cos(x)
        cos_cache[deg] = res
        
        return res


import time
import uuid
import traceback
import random
import trio
import os
import pyglet
import random

from typing import Optional, Callable, Tuple
from .core.entity import Manager, Entity, Component
from .core.vector import Vector
from .core.network import Client, Server
from . import gamedefs
from pyglet.gl import *
from pyglet import clock
import bson



# Credits: trio tutorial
class Tracer(trio.abc.Instrument):
    def before_run(self):
        print("!!! run started")

    def _print_with_task(self, msg, task):
        # repr(task) is perhaps more useful than task.name in general,
        # but in context of a tutorial the extra noise is unhelpful.
        print("{}: {}".format(msg, task.name))

    def task_spawned(self, task):
        self._print_with_task("### new task spawned", task)

    def task_scheduled(self, task):
        self._print_with_task("### task scheduled", task)

    def before_task_step(self, task):
        self._print_with_task(">>> about to run one step of task", task)

    def after_task_step(self, task):
        self._print_with_task("<<< task step finished", task)

    def task_exited(self, task):
        self._print_with_task("### task exited", task)

    def before_io_wait(self, timeout):
        if timeout:
            print("### waiting for I/O for up to {} seconds".format(timeout))
        else:
            print("### doing a quick check for I/O")
        self._sleep_time = trio.current_time()

    def after_io_wait(self, timeout):
        duration = trio.current_time() - self._sleep_time
        print("### finished I/O check (took {} seconds)".format(duration))

    def after_run(self):
        print("!!! run finished")


class EventLoop(object):
    def __init__(self, window):
        self.window = window
        self.stop()

    def run(self):
        trio.run(self.async_run)

    def stop(self):
        self.running = False

    async def async_run(self):
        if self.window:
            self.running = True

            while self.running:
                clock.tick()

                self.window.switch_to()
                self.window.dispatch_events()
                self.window.dispatch_event('on_draw')
                self.window.flip()

                await trio.sleep(0)

            self.window.close()


class Game(object):
    def __init__(self, game_name: str, database_file: Optional[str] = None, client_addr: str = None, server_port: int = None, dedicated: bool = False, player_name: str = None):
        self.dedicated = dedicated or os.environ.get('YODINE_DEDICATED', 'NO').upper() in ('YES', 'TRUE', 'Y')

        self.random = random.Random()
        self.random.seed(time.time())
        
        self.scheduled = []

        server_port = server_port or int(os.environ.get('YODINE_LISTEN', 0)) or None
        client_addr = client_addr or os.environ.get('YODINE_CONNECT', None)

        self.id = str(uuid.uuid4()) # used for network identification
        print("Local ID: ", self.id)

        if not self.dedicated:
            self.window = pyglet.window.Window(visible=False, resizable=True, caption='Yodine')

        else:
            self.window = None

        self.manager = Manager(self, self.window) # type: Manager
        self.database_file = database_file # type: Optional[str]
        self.player_name = player_name or os.environ.get('YODINE_NAME', 'Anon_' + str(random.randint(1, 1999)))

        gamedefs.load_game(game_name, self)
        self.event_loop = EventLoop(self.window)
        self.keyboard = None

        # Enable blending for alpha channel images
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)

        # Multiplayer
        if client_addr:
            client_hostname, client_port = client_addr.split(':')
            self.client = Client(self, (client_hostname, int(client_port)))

        else:
            self.client = None

        if server_port:
            self.server = Server(self, server_port)

        else:
            self.server = None

        # Load the game.
        self.load(game_name)

    def stop(self):
        self.save()

        if self.window:
            self.window.close()
            self.window = None

        if self.event_loop:
            self.event_loop.stop()

        if self.server:
            self.server.stop()

        if self.client:
            self.client.stop()

    async def tick_scheduled(self, dt):
        sc = []

        for s in self.scheduled:
            delay, func, args, kwargs = s[:2]
            delay -= dt

            if delay <= 0:
                func(*args, **kwargs)

            else:
                s[0] = delay
                sc.append(s)

        self.scheduled = sc

    async def loop_schedules(self):
        ot = time.time()
        while True:
            t = time.time()
            dt = t - ot
            ot = t

            await self.tick_scheduled(dt)
            await trio.sleep(1 / 15)

    async def loop(self):
        async with trio.open_nursery() as nursery:
            nursery.start_soon(self.event_loop.async_run)

            if self.server:
                nursery.start_soon(self.server.run)
                self.server.running = True

                if self.dedicated:
                    nursery.start_soon(self.dedicated_ticking)

            if self.client:
                nursery.start_soon(self.client.run)

    async def dedicated_ticking(self):
        ot = time.time()

        while self.server and self.server.running:
            t = time.time()
            dt = t - ot
            ot = t

            self.manager.tick(dt)

            await trio.sleep(1 / 40)

    def load(self, init: Callable):
        for e in self.manager:
            e.remove()

        self.manager.levels = {'_DEFAULT': self.manager.default_level}
        
        if self.database_file and os.path.exists(self.database_file) and not self.client:
            with open(self.database_file, 'rb') as dbfp:
                data = bson.loads(dbfp.read())

                for level in data['levels']:
                    level_obj = self.manager.add_level_save(level['id'], level)

                    for entity in level['entities']:
                        level_obj.load_entity(entity['components'], entity['id'])

                for entity in data['entities']:
                    self.manager.load_entity(entity['components'], entity['id'])

                self.manager.change_level(data['currlevel'])

            found = False

            if not self.dedicated and not self.client:
                for entity in self.manager:
                    if entity.has('name', 'player', 'localplayer'):
                        if entity['name'].value == self.player_name:
                            entity['localplayer'] = self.id # redefine localplayer
                            found = True

                if not found:
                    self.init_player()

            if not len(self.manager):
                gamedefs.init_game(self)

                if not self.dedicated and not self.client:
                    self.init_player()
                
                self.save()

            else:
                self.emit_all('loaded')
            
        else:
            gamedefs.init_game(self)

            if not self.dedicated and not self.client:
                self.init_player()

            self.save()

    def init_player(self):
        for r in self.manager.loader.routines['init.player']:
            r(self)

    def save(self):
        if self.database_file:
            self.emit_all('saved')

            with open(self.database_file, 'wb') as dbfp:
                data = {
                    'levels': [
                        {
                            'id': l.id,
                            'entities': [
                                {
                                    'components': [(c.name, c.value, type(c).__name__) for c in e],
                                    'id': e.id
                                }
                                for e in l.get_entities()
                            ],
                            'deltas': l.deltas
                        } for l in self.manager.levels.values()
                    ],
                    'entities': [
                        {
                            'components': [(c.name, c.value, type(c).__name__) for c in e],
                            'id': e.id
                        }
                        for e in self.manager.get_entities()
                    ],
                    'currlevel': self.manager.current_level.id
                }

                dbfp.write(bson.dumps(data))

    def on_draw(self):
        if not self.dedicated:
            self.manager.render(self.window)

    def on_mouse_drag(self, x, y, dx, dy, button, mod):
        self.emit_all(('mouse', 'drag'), self.window, Vector((x, y)), Vector((dx, dy)), button, mod)

    def on_mouse_motion(self, x, y, dx, dy):
        self.emit_all(('mouse', 'move'), self.window, Vector((x, y)), Vector((dx, dy)))

    def on_mouse_enter(self, x, y):
        self.emit_all(('mouse', 'enter'), self.window, Vector((x, y)))

    def on_mouse_leave(self, x, y):
        self.emit_all(('mouse', 'leave'), self.window, Vector((x, y)))

    def on_mouse_press(self, x, y, button, modifiers):
        self.emit_all(('mouse', 'press'), self.window, Vector((x, y)), button, modifiers)

    def on_mouse_release(self, x, y, button, modifiers):
        self.emit_all(('mouse', 'release'), self.window, Vector((x, y)), button, modifiers)

    def on_key_press(self, button, modifiers):
        self.emit_all(('key', 'press'), self.window, button, modifiers)

    def on_key_release(self, button, modifiers):
        self.emit_all(('key', 'release'), self.window, button, modifiers)

    def emit_all(self, event, *args, **kwargs):
        self.manager.emit_all(event, *args, **kwargs)

    def net_emit(self, source, event, *args, **kwargs):
        if self.client:
            self.client.emit(source, event, *args, **kwargs)
            self.scheduled.append([self.client.average_latency, self.manager.emit, (source, event, *args), kwargs])

        else:
            self.manager.emit(source, event, *args, **kwargs)

    def server_net_emit(self, source, event, *args, **kwargs):
        if self.server:
            self.server.emit(source, event, *args, **kwargs)

    def net_emit_change(self, source, compname, compvalue):
        if self.client:
            self.client.send('change', source, compname, compvalue)

    def on_close(self):
        self.stop()

    def run(self):
        self.keyboard = pyglet.window.key.KeyStateHandler()

        if not self.dedicated:
            self.window.push_handlers(self.keyboard)
            self.window.push_handlers(self)

            def tick(dtime):
                # tick
                t = time.time()
                dt = t - self.last_time
                self.last_time = t

                self.manager.tick(dtime)

            clock.schedule_interval(tick, 1 / 60)

            self.window.set_visible()

        try:
            self.last_time = time.time()
            
            trio.run(self.loop)
            #trio.run(self.loop, instruments=[Tracer()])

        except KeyboardInterrupt:    
            self.stop()
            traceback.print_exc()


import pkg_resources
import os
import random

from tkinter import *
from . import game



class InputRow(object):
    rows = {}

    def __init__(self, master, **kwargs):
        self.master = master
        
        self.row = InputRow.rows.setdefault(master, 0)
        InputRow.rows[master] += 1

        self.init_widgets(**kwargs)

    def init_widgets(self, label=None, initial='', options=None, **_):
        if label is not None:
            self.label_text = StringVar(value=label)
            self.label = Label(self.master, textvariable=self.label_text, bg='white')
            self.label.grid(row=self.row, column=0, sticky='e')

        if options:
            self.value = StringVar()
            self.value.set(initial)
            self.input = OptionMenu(self.master, self.value, *options)

        else:
            self.value = StringVar(value=initial)
            self.input = Entry(self.master, textvariable=self.value)

        self.input.grid(row=self.row, column=1, sticky='nswe')


class LauncherArea(Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, bd=2, relief=SUNKEN, **kwargs)

        self.init_widgets()

        for child in self.winfo_children():
            child.grid_configure(padx=5, pady=5)

    def init_widgets(self):
        pass


class LauncherInputArea(LauncherArea):
    def init_widgets(self):
        self.games = []

        for ep in pkg_resources.iter_entry_points('yodine.plugin'):
            self.games.append(ep.name)

        self.games = list(set(self.games))

        if 'yodine_data' in self.games:
            init_game = 'yodine_data'

        else:
            init_game = random.choice(self.games)

        self.player_name = InputRow(self, label='Player Name*', initial=os.environ.get('YODINE_NAME', None)).value
        self.game_name = InputRow(self, label='Game', options=self.games, initial=init_game).value
        self.save_name = InputRow(self, label='Save Filename*').value
        self.server_addr = InputRow(self, label='Connect to*', initial=os.environ.get('YODINE_CONNECT', None)).value

        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=3)
        
        notice = Label(self, text='* optional', justify=CENTER, bg='white')
        notice.grid(column=0, columnspan=2, sticky='we')


class YodineLauncher(Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.grid(column=0, row=0, sticky='nswe')

        self.init_widgets()

        for child in self.winfo_children():
            child.grid_configure(padx=5, pady=5)

    def init_widgets(self):
        self.title = Label(self, text="Yodine Launcher", bg='white')
        self.title.grid(row=0)

        self.form = LauncherInputArea(self, bg='white')
        self.form.grid(row=1, sticky='nswe')

        self.start_button = Button(self, command=self.on_start, text="Run")
        self.start_button.grid(row=2, sticky='we')

    def on_start(self):
        self.master.destroy()
        run_game(self.form.game_name.get(), self.form.save_name.get(), self.form.server_addr.get(), self.form.player_name.get())


def run_game(game_name, save_name = None, connect_to = None, player_name = None):
    this_game = game.Game(game_name, save_name or None, client_addr=connect_to or None, player_name=player_name or None)
    print('Starting...')
    this_game.run()


def main():
    root = Tk()
    root.title('Yodine Launcher')

    frame = YodineLauncher(root, background='white')
    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)

    root.mainloop()



if __name__ == '__main__':
    main()


import sys
import os


default_module = """
# ============== {name} ==============
#             by {author}
# licensed under {license}
#
#   {description}

# This is the main code of your plugin.
# Here you will add all of the logic that will
# be used by your plugin.

# The lines below import all of the classes you'll
# use from Yodine.
from yodine.core.entity import EntityTemplate, System, TileType
from yodine.core.extension import ModLoader
from yodine.core.vector import ComponentVector

# Other general imports below.
import pyglet
import os



# Defines an extra resource path.
pyglet.resource.path.append(os.path.join(os.path.split(__file__)[0], 'assets'))
pyglet.resource.reindex()


# These helper functions will access assets for you :)
def asset_path(asset_name: str) -> str:
    return os.path.join(os.path.split(__file__)[0], 'assets', asset_name)

def open_asset(asset_name: str) -> file:
    return open(asset_path(asset_name))


# This function will be called when the plugin is
# loaded.
def loaded(loader: ModLoader):
    # Defines a standard tile type (background - does nothing).
    class FloorTileType(TileType):
        name = 'floor'

    loader.add_tile_type(FloorTileType(pyglet.resource.image(asset_path('tiles/floor.png'))))

    # Defines a custom tile type (foreground - e.g. collides).
    class WallTileType(TileType):
        name = 'wall'

        def on_move(self, manager, entity, start_pos):
            if self.is_inside(manager, entity):
                ComponentVector(entity['position']) << start_pos

    loader.add_tile_type(WallTileType(pyglet.resource.image(asset_path('tiles/wall.png'))))
    
    # Defines a routine, which may be used by this or other
    # plugins.
    @loader.routine()
    def moo():
        # This can later be accessed via loader.routine (see
        # MooSystem).
        with open_asset('moo.txt') as moofile:
            print(moofile.read().strip())

    # Defines a game initialization routine, by giving the routine decorator a group string.
    @loader.routine('yodine.init')
    def begin():
        print("Welcome to the Moo Moo Mountains!")

    # Defines a system to be registered by
    # the loader.
    @loader.system_type
    class MyPluginSystem(System):
        # This is a list of components that the entity MUST
        # have, and that will be passed to tick, render, and
        # to event handlers (on_*).
        component_types = ['name']

        # This is a list of components that the entity MAY
        # have, and that will be passed to tick, render, and
        # to event handlers (on_*).
        # -----
        # If those components are not present, the default values
        # are passed instead.
        component_defaults = {{'postfix': "I'm a dumb plugin, teehee!"}}

        # This method of a system is called every tick, i.e., everytime
        # the game updates itself.
        # It is responsible for updating the entity.
        def tick(self, entity: Entity, *args, **kwargs) -> None:
            print("TICK!")

        # This method of a system is called every time the game is rendered.
        # It is responsible for rendering the entity.
        def render(self, entity: Entity, window: pyglet.window.Window, name, postfix, *args, **kwargs) -> None:
            print("I will print {{}}... already did! {{}}".format(name, postfix))
    
    @loader.system_type
    class MooSystem(System):
        # Components that must have certain values
        # in order for this system to operate on
        # a certain entity. Note that 'operate' includes
        # `tick`, `render` and `on_*` system methods.
        component_checks = {{'moo': 'YES!'}}

        def render(self, entity: Entity, *args, **kwargs) -> None:
            # Grabs a routine and calls it.
            loader.routines.moo()
            
            # This is our moo() routine defined earlier!
            # How magic is that? :)

    # Defines an entity template, a way to specify how certian
    # entities should be created.
    @loader.template
    class MyPluginEntityType(EntityTemplate):
        # The name of this template. Used when looking it up.
        name = 'Cow'

        # The group of this template. Used when looking a specific group of templates up.
        group = 'living.passive'

        # A list of default components.
        default_components = [
            ('name', 'a cow'),
            ('postfix', 'MOOO!'),
            ('moo', 'YES!')
        ]


""".lstrip()

setup_template = """
from setuptools import setup
from {name}.version import VERSION_STR as VERSION

import os



NAME = '{name}'
DESCRIPTION = '{description}'
AUTHOR = '{author}'
REQUIRES_PYTHON = '>=3.5.0'
LICENSE = '{license}'


# What packages are required for this plugin to work?
REQUIRED = [
    # Yodine plugin dependencies:
    'yodine_data',

    # General dependencies:
    # 'numpy'
]

with open('README.md') as readme:
    long_description = readme.read()
        

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    python_requires=REQUIRES_PYTHON,
    packages=['{name}'],

    entry_points={{
        'yodine.plugin': ['{{0}} = {{0}}.plugin:loaded'.format(NAME)],
    }},
    install_requires=REQUIRED,
    license=LICENSE,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    package_data={{NAME: ['assets/*', 'assets/**/*']}},
    include_package_data=True,
)
""".lstrip()

gitignore = """
# === Python ===

# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions
*.so

# Distribution / packaging
.Python
env/
fastenv/
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
*.egg-info/
.installed.cfg
*.egg

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*,cover
.hypothesis/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py

# Sphinx documentation
docs/_build/

# PyBuilder
target/

#Ipython Notebook
.ipynb_checkpoints

# pyenv
.python-version
""".lstrip()

readme_template = """
# {name}

**A Yodine plugin by {author}**

{description}

-----

Licensed under {license}.
""".lstrip()


def init_plugin(plugin_name, folder='.', author='an anonymous pluginwriter', description='', license='MIT'):
    plugin_pkg = os.path.join(folder, plugin_name)
    plugin_init = os.path.join(plugin_pkg, '__init__.py')
    plugin_code = os.path.join(plugin_pkg, 'plugin.py')
    asset_folder = os.path.join(plugin_pkg, 'assets')
    setup_py = os.path.join(folder, 'setup.py')
    setup_cfg = os.path.join(folder, 'setup.cfg')
    readme = os.path.join(folder, 'README.md')
    manifest = os.path.join(folder, 'MANIFEST.in')
    gitignore_path = os.path.join(folder, '.gitignore')
    version = os.path.join(plugin_pkg, 'version.py')

    if not os.path.exists(plugin_pkg):
        os.makedirs(plugin_pkg)

    if not os.path.exists(plugin_code):
        open(plugin_code, 'w').write(default_module.format(name=plugin_name, author=author, description=description, license=license))

    if not os.path.exists(plugin_init):
        open(plugin_init, 'w')

    if not os.path.exists(asset_folder):
        os.makedirs(asset_folder)

    if not os.path.exists(setup_py):
        open(setup_py, 'w').write(setup_template.format(name=plugin_name, author=author.replace('\'', '\\\''), description=description.replace('\'', '\\\''), license=license.replace('\'', '\\\'')))

    if not os.path.exists(readme):
        open(readme, 'w').write(readme_template.format(name=plugin_name, author=author, description=description, license=license))

    if not os.path.exists(gitignore_path):
        open(gitignore_path, 'w').write(gitignore)

    if not os.path.exists(version):
        open(version, 'w').write('VERSION_TUPLE = (0, 1, 0)\nVERSION_POSTFIX = \'\'\n\nVERSION_STR = \'.\'.join(str(v) for v in VERSION_TUPLE) + VERSION_POSTFIX\n')


if __name__ == "__main__":
    print("What is the name of your brand new plugin?")
    plugin_name = input('> ')
    print("What is your author name? (can be your real name, a nickname, or even just your email!)")
    author = input('> ')
    print("Give your plugin a brief description.")
    description = input('> ')
    print("And a license! (default: MIT)")
    license = input('> ') or 'MIT'
    print("In which folder do you want to set this plugin project up? (default: {})".format(plugin_name))
    folder = input('> ') or plugin_name
    print()
    
    try:
        init_plugin(plugin_name, folder, author, description, license)

    except RuntimeError:
        print("Plugin already found (setup.py), aborting.")
        exit(1)

    else:
        print("Plugin project initialized!")
        exit(0)